using System;
using System.IO;
using System.Collections;
using System.ComponentModel;
using System.Drawing;
using System.Data;
using System.Windows.Forms;
using System.Drawing.Imaging;
using System.Runtime.InteropServices;

// Interop
using AxSHDocVw;
using mshtml;
using SHDocVw;
using FreeImageAPI;

namespace WebsiteCapture
{
    /// <summary>
    /// Converts HTML documents to PNG format.
    /// </summary>
    public class Html2Image : System.Windows.Forms.UserControl
    {
        private AxSHDocVw.AxWebBrowser browser;
        private System.ComponentModel.Container m_Components = null;

        private Html2Image()
        {
            InitializeComponent();

            // Go to the blank page, initializes the WebBrowser control to handle HTML documents.
            browser.Navigate("about:blank");

            // Wait for the control to initialize to HTML and load the blank page.
            while (body == null)
            {
                Application.DoEvents();
            }
            browser.DocumentComplete += new AxSHDocVw.DWebBrowserEvents2_DocumentCompleteEventHandler(browser_DocumentComplete);
        }

        private bool m_DocComplete = false;

        #region Public HTML Properties

        [Browsable(false)]  // Do not show in the properties box
        public HTMLDocument document
        {
            get
            {
                return (HTMLDocument)browser.Document;
            }
        }

        [Browsable(false)]  // Do not show in the properties box
        public HTMLBody body
        {
            get
            {
                return (HTMLBody)document.body;
            }
        }

        // Document Body's innerText
        [DefaultValue("")] // Let the properties box know the default value
        [Browsable(true)]  // Do show in the properties box
        [Description("Internet Explorer's BODY.innerText Property")]
        public override string Text
        {
            get { return body.innerText; }
            set { body.innerText = value; }
        }

        // Document Body's innerHTML
        [DefaultValue("")]
        [Description("Internet Explorer's BODY.innerHTML Property")]
        public string HTML
        {
            get { return body.innerHTML; }
            set { body.innerHTML = value; }
        }

        #endregion

        /// <summary>
        /// This function downloads a whole site and saves it to a bitmap
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        void browser_DocumentComplete(object sender, DWebBrowserEvents2_DocumentCompleteEvent e)
        {
            SHDocVw.IWebBrowser2 doc = e.pDisp as SHDocVw.IWebBrowser2;
            if (doc != (sender as AxSHDocVw.AxWebBrowser).GetOcx())
                return; //still loading other frames

            document.body.setAttribute("scroll", "no", 0);
            hideBorder();

            Bitmap bmp = CaptureImage();
            if (resize != 100)
                bmp = ResizeImage(bmp, (int)(bmp.Width * resize / 100), (int)(bmp.Height * resize / 100));
            saveBitmap(bmp, outputPath);
            this.m_DocComplete = true;
        }

        /// <summary>
        /// Converts an HTML file to bitmap
        /// </summary>
        /// <param name="url">A string with the url of the HTML file that should be converted.</param>
        public static void ConvertPage(string url, string outputPath, int resize)
        {
            Html2Image html2Image = new Html2Image();
            html2Image.CreateControl();
            html2Image.SavePage(url, outputPath, resize);
            html2Image.Dispose();
        }

        /// <summary>
        /// Converts an html fragment to bitmap
        /// </summary>
        /// <param name="html"></param>
        /// <param name="outputPath"></param>
        public static void ConvertHtml(string html, string outputPath, int resize)
        {
            Html2Image html2Image = new Html2Image();
            html2Image.CreateControl();
            html2Image.SaveHtml(html, outputPath, resize);
            html2Image.Dispose();
        }

        string outputPath;
        int resize;

        /// <summary>
        /// Saves a complete page to bitmap
        /// </summary>
        /// <param name="url"></param>
        /// <param name="outputPath"></param>
        public void SavePage(string url, string outputPath, int resize)
        {
            this.outputPath = outputPath;
            this.resize = resize;

            while (!browser.Created)
            {
                Application.DoEvents();
            }

            object _arg1 = 0;
            object missing = "";
            browser.Navigate(url, ref _arg1, ref missing, ref missing, ref missing);

            while (!this.m_DocComplete)
            {
                Application.DoEvents();
            }
        }

        /// <summary>
        /// Saves an html fragment to bitmap
        /// </summary>
        /// <param name="html"></param>
        /// <param name="outputPath"></param>
        public void SaveHtml(string html, string outputPath, int resize)
        {
            this.outputPath = outputPath;
            this.resize = resize;

            while (!browser.Created)
            {
                Application.DoEvents();
            }

            object _arg1 = 0;
            object missing = "";
            this.HTML = html;

            hideBorder();
            
            Bitmap bmp = CaptureImage();
            if(resize != 100)
                bmp = ResizeImage(bmp, (int)(bmp.Width  * resize / 100), (int)(bmp.Height * resize / 100));

            saveBitmap(bmp, outputPath);
        }

        private void hideBorder()
        {
            document.body.style.border = "none"; //remove border
        }

        private void saveBitmap(Bitmap bmp, string outputPath)
        {
            string ext = Path.GetExtension(outputPath);
            switch (ext)
            {
                case ".tga":
                    {
                        string tmpPath = Path.ChangeExtension(outputPath, ".png");
                        bmp.Save(tmpPath);
                        uint imgPtr = FreeImage.Load(FREE_IMAGE_FORMAT.FIF_PNG, tmpPath, 0);
                        FreeImage.Save(FREE_IMAGE_FORMAT.FIF_TARGA, imgPtr, outputPath, 0);
                        File.Delete(tmpPath);
                    }
                    break;

                default:
                    bmp.Save(outputPath);
                    break;
            }                
        }

        private Bitmap ResizeImage(Bitmap b, int nWidth, int nHeight)
        {
            try
            {
                Bitmap result = new Bitmap(nWidth, nHeight);
                using (Graphics g = Graphics.FromImage((Image)result))
                    g.DrawImage(b, 0, 0, nWidth, nHeight);
                return result;
            }
            catch (ArgumentException)
            {
                return b;
            }
        }

        private Bitmap CaptureImage()
        {
            try
            {
                int htmlWidth  = this.body.scrollWidth;
                int htmlHeight = this.body.scrollHeight;

                int verticalScrollWidth = Convert.ToInt32(this.body.rightMargin) + Convert.ToInt32(this.body.leftMargin);
                int offW = 60, offH = 30;

                // Make our embedded ie control that size
                if ((browser.Height != htmlHeight + offH) || (browser.Width != htmlWidth + offW))
                {
                    browser.Height = htmlHeight + offH;
                    browser.Width = htmlWidth + offW;
                }

                // Setup bitmap memory and wrap a DC around it
                Bitmap _bitmap = new Bitmap(htmlWidth + offW - verticalScrollWidth, htmlHeight + offH, PixelFormat.Format24bppRgb);
                Graphics _graphics = Graphics.FromImage(_bitmap);
                IntPtr _memdc = _graphics.GetHdc();
                IntPtr _hbitmap = _bitmap.GetHbitmap();
                SelectObject(_memdc, _hbitmap);

                // Tell ie to print into our dc/bitmap

                // Use PrintWindow windows api
                //
                const uint PW_CLIENTONLY = 0x00000001;
                bool _rv = PrintWindow(browser.Handle, _memdc, PW_CLIENTONLY);

                // Send it a print msg + flags
                const uint WM_PRINT = 0x0317;
                // const uint WM_PRINTCLIENT = 0x0318;
                // const uint PRF_CHECKVISIBLE = 0x00000001;
                const uint PRF_NONCLIENT = 0x00000002;
                const uint PRF_CLIENT = 0x00000004;
                const uint PRF_ERASEBKGND = 0x00000008;
                const uint PRF_CHILDREN = 0x00000010;
                const uint PRF_OWNED = 0x00000020;
                int _err = SendMessage(browser.Handle, WM_PRINT, (uint)_memdc, (uint)
                    (PRF_CLIENT | PRF_NONCLIENT | PRF_OWNED | PRF_CHILDREN | PRF_ERASEBKGND));

                // Save image
                Bitmap _bitmap2 = Bitmap.FromHbitmap(_hbitmap);

                // Cleanup
                DeleteObject(_hbitmap);
                _graphics.ReleaseHdc(_memdc);
                _graphics.Dispose();
                _bitmap.Dispose();

                return _bitmap2;

            }
            catch (Exception ex)
            {
                Console.WriteLine(ex.ToString());
                throw;
            }
        }

        protected override void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (m_Components != null)
                {
                    m_Components.Dispose();
                }
            }
            base.Dispose(disposing);
        }

        private void InitializeComponent()
        {
            this.browser = new AxSHDocVw.AxWebBrowser();
            ((System.ComponentModel.ISupportInitialize)(this.browser)).BeginInit();
            this.SuspendLayout();
            //
            // m_IE
            //
            this.browser.Enabled = true;
            this.browser.Location = new System.Drawing.Point(0, 0);
            this.browser.Size = new System.Drawing.Size(804, 580);
            

            //
            // Html2Image
            //
            this.Controls.AddRange(new System.Windows.Forms.Control[] { this.browser });
            this.Size = new System.Drawing.Size(292, 266);
            ((System.ComponentModel.ISupportInitialize)(this.browser)).EndInit();
            this.ResumeLayout(false);
        }

        #region Win32 InterOp
        /// <summary>
        /// 
        /// </summary>
        /// <param name="hdc">handle to DC</param>
        /// <param name="hgdiobj">handle to object</param>
        /// <returns></returns>
        [DllImport("Gdi32.dll")]
        private static extern IntPtr SelectObject(IntPtr hdc, IntPtr hgdiobj);

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hObject">handle to graphic object</param>
        /// <returns></returns>
        [DllImport("Gdi32.dll")]
        private static extern bool DeleteObject(IntPtr hObject);

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hwnd">Window to copy</param>
        /// <param name="hdcBlt">HDC to print into</param>
        /// <param name="nFlags">Optional flags // must contain PW_CLIENTONLY</param>
        /// <returns></returns>
        [DllImport("User32.dll")]
        private static extern bool PrintWindow(IntPtr hwnd, IntPtr hdcBlt, uint nFlags);

        /// <summary>
        /// 
        /// </summary>
        /// <param name="hWnd">handle to destination window</param>
        /// <param name="Msg">message</param>
        /// <param name="wParam">firstmessage parameter</param>
        /// <param name="lParam">second message parameter</param>
        /// <returns></returns>
        [DllImport("User32.dll")]
        private static extern int SendMessage(IntPtr hWnd, uint Msg, uint wParam, uint lParam);

        #endregion
    }
}