using System;
using System.Drawing;
using System.Runtime.InteropServices;
using System.Collections;
using System.Drawing.Imaging;
using System.IO;
using System.Windows.Forms;

namespace Common
{
    public enum ListDisplayMode
    {
        Thumbnails,
        Tiles,
        Icons,
        Details
    }

    public enum EIconSize
    {
        Small = 0, /// 32x32
        Large = 1  /// 16x16
    }

    ///
    /// reads icons from the windows shell
    ///
    /// state of the folder
    public enum EnumFolderType
    {
        Open = 0, /// open folder
        Closed = 1 /// closed folder
    }

    public class IconReader
    {
        static Hashtable htSmallPictureIcons = new Hashtable();
        static Hashtable htLargePictureIcons = new Hashtable();

        ///
        /// used to get/return information about a file
        /// in SHGetFileInfo call
        ///
        [System.FlagsAttribute()]
        private enum EnumFileInfoFlags : uint
        {
            LARGEICON = 0x000000000, /// get large icon
            SMALLICON = 0x000000001, /// get small icon
            OPENICON = 0x000000002, /// get open icon
            SHELLICONSIZE = 0x000000004, /// get shell size icon
            PIDL = 0x000000008, /// pszPath is a pidl
            USEFILEATTRIBUTES = 0x000000010, /// use passed dwFileAttribute
            ADDOVERLAYS = 0x000000020, /// apply the appropriate overlays
            OVERLAYINDEX = 0x000000040, /// get the index of the overlay
            ICON = 0x000000100, /// get icon
            DISPLAYNAME = 0x000000200, /// get display name
            TYPENAME = 0x000000400, /// get type name
            ATTRIBUTES = 0x000000800, /// get attributes
            ICONLOCATION = 0x000001000, /// get icon location
            EXETYPE = 0x000002000, /// return exe type
            SYSICONINDEX = 0x000004000, /// get system icon index
            LINKOVERLAY = 0x000008000, /// put a link overlay on icon
            SELECTED = 0x000010000, /// show icon in selected state
            ATTR_SPECIFIED = 0x000020000 /// get only specified attributes
        }

        /// maxumum length of path
        private const int conMAX_PATH = 260;

        /// looking for folder
        private const uint conFILE_ATTRIBUTE_DIRECTORY = 0x00000010;

        /// looking for file
        private const uint conFILE_ATTRIBUTE_NORMAL = 0x00000080;

        ///
        /// hold file/icon information
        /// see platformSDK SHFILEINFO
        ///
        ///
        /// be sure to call DestroyIcon [hIcon] when done
        ///
        [StructLayout(LayoutKind.Sequential)]
        private struct ShellFileInfo
        {
            public const int conNameSize = 80;
            public System.IntPtr hIcon;  // note to call DestroyIcon
            public int iIndex;
            public uint dwAttributes;

            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = conMAX_PATH)]
            public string szDisplayName;

            [MarshalAs(UnmanagedType.ByValTStr, SizeConst = conNameSize)]
            public string szTypeName;
        };

        /// used to free a windows icon handle
        ///
        /// "hIcon">icon handle.
        [DllImport("User32.dll")]
        private static extern int DestroyIcon(System.IntPtr hIcon);

        /// gets file information
        /// see platformSDK
        [DllImport("Shell32.dll")]
        private static extern System.IntPtr SHGetFileInfo(
          string pszPath,
          uint dwFileAttributes,
          ref ShellFileInfo psfi,
          uint cbFileInfo,
          uint uFlags
        );



        /// <summary>
        /// lookup and return an icon from windows shell
        /// </summary>
        /// <param name="filePath">path to the file</param>
        /// <param name="size">large or small</param>
        /// <param name="addLinkOverlay">true to include the overlay link iconlet</param>
        /// <returns>requested icon</returns>
        public static Icon GetFileIcon(string filePath, EIconSize size)
        {
            EnumFileInfoFlags flags = EnumFileInfoFlags.TYPENAME;
            ShellFileInfo shellFileInfo = new ShellFileInfo();

            SHGetFileInfo(
              filePath,
              conFILE_ATTRIBUTE_NORMAL,
              ref shellFileInfo,
              (uint)System.Runtime.InteropServices.Marshal.SizeOf(shellFileInfo),
              (uint)flags);

            string typeName = shellFileInfo.szTypeName;
            Icon icon;
            if (size == EIconSize.Small)
            {
                object obj = htSmallPictureIcons[typeName];
                if (obj != null)
                    return (Icon)obj;
                flags = EnumFileInfoFlags.ICON | EnumFileInfoFlags.SMALLICON;
                shellFileInfo = new ShellFileInfo();
                SHGetFileInfo(
                  filePath,
                  conFILE_ATTRIBUTE_NORMAL,
                  ref shellFileInfo,
                  (uint)Marshal.SizeOf(shellFileInfo),
                  (uint)flags);
                // deep copy
                icon = (Icon)Icon.FromHandle(shellFileInfo.hIcon).Clone();
                htSmallPictureIcons[typeName] = icon;
            }
            else
            {
                object obj = htLargePictureIcons[typeName];
                if (obj != null)
                    return (Icon)obj;
                flags = EnumFileInfoFlags.ICON | EnumFileInfoFlags.LARGEICON;
                shellFileInfo = new ShellFileInfo();
                SHGetFileInfo(
                  filePath,
                  conFILE_ATTRIBUTE_NORMAL,
                  ref shellFileInfo,
                  (uint)Marshal.SizeOf(shellFileInfo),
                  (uint)flags);
                // deep copy
                icon = (Icon)Icon.FromHandle(shellFileInfo.hIcon).Clone();
                htLargePictureIcons[typeName] = icon;
            }

            // release handle
            DestroyIcon(shellFileInfo.hIcon);
            return icon;
        }

        /// <summary>
        /// lookup and return an icon from windows shell
        /// </summary>
        /// <param name="size">large or small</param>
        /// <param name="folderType">open or closed</param>
        /// <returns>requested icon</returns>
        public static Icon GetFolderIcon(EIconSize size, EnumFolderType folderType)
        {
            return GetFolderIcon(null, size, folderType);
        }

        /// <summary>
        /// lookup and return an icon from windows shell
        /// </summary>
        /// <param name="folderPath">path to folder</param>
        /// <param name="size">large or small</param>
        /// <param name="folderType">open or closed</param>
        /// <returns>requested icon</returns>
        public static Icon GetFolderIcon(string folderPath, EIconSize size, EnumFolderType folderType)
        {
            EnumFileInfoFlags flags = EnumFileInfoFlags.ICON | EnumFileInfoFlags.USEFILEATTRIBUTES;

            if (folderType == EnumFolderType.Open)
                flags |= EnumFileInfoFlags.OPENICON;

            if (EIconSize.Small == size)
                flags |= EnumFileInfoFlags.SMALLICON;
            else
                flags |= EnumFileInfoFlags.LARGEICON;

            ShellFileInfo shellFileInfo = new ShellFileInfo();
            SHGetFileInfo(
              folderPath,
              conFILE_ATTRIBUTE_DIRECTORY,
              ref shellFileInfo,
              (uint)System.Runtime.InteropServices.Marshal.SizeOf(shellFileInfo),
              (uint)flags);

            // deep copy
            Icon icon = (Icon)Icon.FromHandle(shellFileInfo.hIcon).Clone();

            // release handle
            DestroyIcon(shellFileInfo.hIcon);
            return icon;
        }
    }

    public class IconManager
    {
        static Hashtable htSmallIcons = new Hashtable();
        static ImageList imageListSmallIcons = new ImageList();
        static Hashtable htLargeIcons = new Hashtable();
        static ImageList imageListLargeIcons = new ImageList();

        public static int GetImageIndex(string path, EIconSize iconsize)
        {
            Icon icon = IconReader.GetFileIcon(path, iconsize);
            Hashtable htIcons = (iconsize == EIconSize.Small) ? htSmallIcons : htLargeIcons;
            ImageList imageListIcons = (iconsize == EIconSize.Small) ? imageListSmallIcons : imageListLargeIcons;
            if (htIcons[icon] == null)
            {
                imageListIcons.Images.Add(icon);
                htIcons[icon] = imageListIcons.Images.Count - 1;
            }
            return (int)htIcons[icon];
        }

        public ImageList SmallIcons
        {
            get
            {
                return imageListSmallIcons;
            }
        }

        public ImageList LargeIcons
        {
            get
            {
                return imageListLargeIcons;
            }
        }
    }
}